@@ -1,5 +1,5 @@
 /*
- * Copyright © 2020-2021 Inria.  All rights reserved.
+ * Copyright © 2020-2022 Inria.  All rights reserved.
  * See COPYING in top-level directory.
  */
 
@@ -16,12 +16,14 @@
 
 static void
 hwloc__levelzero_properties_get(ze_device_handle_t h, hwloc_obj_t osdev,
-                                int sysman_maybe_missing)
+                                int sysman_maybe_missing,
+                                int *is_integrated_p)
 {
   ze_result_t res;
   ze_device_properties_t prop;
   zes_device_properties_t prop2;
   int is_subdevice = 0;
+  int is_integrated = 0;
 
   memset(&prop, 0, sizeof(prop));
   res = zeDeviceGetProperties(h, &prop);
@@ -30,7 +32,10 @@ hwloc__levelzero_properties_get(ze_device_handle_t h, hwloc_obj_t osdev,
      * flags 1<<0 means integrated (vs discrete).
      */
     char tmp[64];
+    char uuid[ZE_MAX_DEVICE_UUID_SIZE*2+1];
+    unsigned i;
     const char *type;
+
     switch (prop.type) {
     case ZE_DEVICE_TYPE_GPU: type = "GPU"; break;
     case ZE_DEVICE_TYPE_CPU: type = "CPU"; break;
@@ -52,10 +57,20 @@ hwloc__levelzero_properties_get(ze_device_handle_t h, hwloc_obj_t osdev,
     snprintf(tmp, sizeof(tmp), "%u", prop.numThreadsPerEU);
     hwloc_obj_add_info(osdev, "LevelZeroNumThreadsPerEU", tmp);
 
+    for(i=0; i<ZE_MAX_DEVICE_UUID_SIZE; i++)
+      snprintf(uuid+2*i, 3, "%02x", prop.uuid.id[i]);
+    hwloc_obj_add_info(osdev, "LevelZeroUUID", uuid);
+
     if (prop.flags & ZE_DEVICE_PROPERTY_FLAG_SUBDEVICE)
       is_subdevice = 1;
+
+    if (prop.flags & ZE_DEVICE_PROPERTY_FLAG_INTEGRATED)
+      is_integrated = 1;
   }
 
+  if (is_integrated_p)
+    *is_integrated_p = is_integrated;
+
   if (is_subdevice)
     /* sysman API on subdevice returns the same as root device, and we don't need those duplicate attributes */
     return;
@@ -229,7 +244,8 @@ hwloc__levelzero_memory_get_from_sysman(zes_device_handle_t h,
 
 static void
 hwloc__levelzero_memory_get_from_coreapi(ze_device_handle_t h,
-                                         hwloc_obj_t osdev)
+                                         hwloc_obj_t osdev,
+                                         int ignore_ddr)
 {
   ze_device_memory_properties_t *mh;
   uint32_t nr_mems;
@@ -257,6 +273,8 @@ hwloc__levelzero_memory_get_from_coreapi(ze_device_handle_t h,
                     mh[m].name, (unsigned long long) mh[m].totalSize, osdev->name);
         if (!mh[m].totalSize)
           continue;
+        if (ignore_ddr && !strcmp(_name, "DDR"))
+          continue;
         if (!_name[0])
           _name = "Memory";
         snprintf(name, sizeof(name), "LevelZero%sSize", _name); /* HBM or DDR, or Memory if unknown */
@@ -270,7 +288,7 @@ hwloc__levelzero_memory_get_from_coreapi(ze_device_handle_t h,
 
 
 static void
-hwloc__levelzero_memory_get(zes_device_handle_t h, hwloc_obj_t root_osdev,
+hwloc__levelzero_memory_get(zes_device_handle_t h, hwloc_obj_t root_osdev, int is_integrated,
                             unsigned nr_subdevices, zes_device_handle_t *subh, hwloc_obj_t *sub_osdevs)
 {
   static int memory_from_coreapi = -1; /* 1 means coreapi, 0 means sysman, -1 means sysman if available or coreapi otherwise */
@@ -298,11 +316,12 @@ hwloc__levelzero_memory_get(zes_device_handle_t h, hwloc_obj_t root_osdev,
     first = 0;
   }
 
-  if (memory_from_coreapi == 1) {
+  if (memory_from_coreapi > 0) {
     unsigned k;
-    hwloc__levelzero_memory_get_from_coreapi(h, root_osdev);
+    int ignore_ddr = (memory_from_coreapi != 2) && is_integrated; /* DDR ignored in integrated GPUs, it's like the host DRAM */
+    hwloc__levelzero_memory_get_from_coreapi(h, root_osdev, ignore_ddr);
     for(k=0; k<nr_subdevices; k++)
-      hwloc__levelzero_memory_get_from_coreapi(subh[k], sub_osdevs[k]);
+      hwloc__levelzero_memory_get_from_coreapi(subh[k], sub_osdevs[k], ignore_ddr);
   } else {
     hwloc__levelzero_memory_get_from_sysman(h, root_osdev, nr_subdevices, sub_osdevs);
     /* no need to call hwloc__levelzero_memory_get() on subdevices,
@@ -342,7 +361,12 @@ hwloc_levelzero_discover(struct hwloc_backend *backend, struct hwloc_disc_status
    */
   env = getenv("ZES_ENABLE_SYSMAN");
   if (!env) {
-    putenv((char *) "ZES_ENABLE_SYSMAN=1");
+    /* setenv() is safer than putenv() but not available on Windows */
+#ifdef HWLOC_WIN_SYS
+    putenv("ZES_ENABLE_SYSMAN=1");
+#else
+    setenv("ZES_ENABLE_SYSMAN", "1", 1);
+#endif
     /* we'll warn in hwloc__levelzero_properties_get() if we fail to get zes devices */
     sysman_maybe_missing = 1;
   } else if (!atoi(env)) {
@@ -352,7 +376,7 @@ hwloc_levelzero_discover(struct hwloc_backend *backend, struct hwloc_disc_status
   res = zeInit(0);
   if (res != ZE_RESULT_SUCCESS) {
     if (!hwloc_hide_errors()) {
-      fprintf(stderr, "Failed to initialize LevelZero in ze_init(): %d\n", (int)res);
+      fprintf(stderr, "Failed to initialize LevelZero in ze_init(): 0x%x\n", (unsigned)res);
     }
     return 0;
   }
@@ -361,6 +385,7 @@ hwloc_levelzero_discover(struct hwloc_backend *backend, struct hwloc_disc_status
   res = zeDriverGet(&nbdrivers, NULL);
   if (res != ZE_RESULT_SUCCESS || !nbdrivers)
     return 0;
+  hwloc_debug("hwloc/L0: found %u drivers\n", (unsigned) nbdrivers);
   drh = malloc(nbdrivers * sizeof(*drh));
   if (!drh)
     return 0;
@@ -380,6 +405,7 @@ hwloc_levelzero_discover(struct hwloc_backend *backend, struct hwloc_disc_status
     res = zeDeviceGet(drh[i], &nbdevices, NULL);
     if (res != ZE_RESULT_SUCCESS || !nbdevices)
       continue;
+    hwloc_debug("hwloc/L0: found %u devices in driver #%u\n", (unsigned) nbdevices, (unsigned) i);
     dvh = malloc(nbdevices * sizeof(*dvh));
     if (!dvh)
       continue;
@@ -390,11 +416,11 @@ hwloc_levelzero_discover(struct hwloc_backend *backend, struct hwloc_disc_status
     }
 
     for(j=0; j<nbdevices; j++) {
-      zes_pci_properties_t pci;
       zes_device_handle_t sdvh = dvh[j];
       zes_device_handle_t *subh = NULL;
       uint32_t nr_subdevices;
       hwloc_obj_t osdev, parent, *subosdevs = NULL;
+      int is_integrated = 0;
 
       osdev = hwloc_alloc_setup_object(topology, HWLOC_OBJ_OS_DEVICE, HWLOC_UNKNOWN_INDEX);
       snprintf(buffer, sizeof(buffer), "ze%u", zeidx); // ze0d0 ?
@@ -409,7 +435,7 @@ hwloc_levelzero_discover(struct hwloc_backend *backend, struct hwloc_disc_status
       snprintf(buffer, sizeof(buffer), "%u", j);
       hwloc_obj_add_info(osdev, "LevelZeroDriverDeviceIndex", buffer);
 
-      hwloc__levelzero_properties_get(dvh[j], osdev, sysman_maybe_missing);
+      hwloc__levelzero_properties_get(dvh[j], osdev, sysman_maybe_missing, &is_integrated);
 
       hwloc__levelzero_cqprops_get(dvh[j], osdev);
 
@@ -435,7 +461,7 @@ hwloc_levelzero_discover(struct hwloc_backend *backend, struct hwloc_disc_status
             snprintf(tmp, sizeof(tmp), "%u", k);
             hwloc_obj_add_info(subosdevs[k], "LevelZeroSubdeviceID", tmp);
 
-            hwloc__levelzero_properties_get(subh[j], subosdevs[k], sysman_maybe_missing);
+            hwloc__levelzero_properties_get(subh[j], subosdevs[k], sysman_maybe_missing, NULL);
 
             hwloc__levelzero_cqprops_get(subh[k], subosdevs[k]);
           }
@@ -448,19 +474,42 @@ hwloc_levelzero_discover(struct hwloc_backend *backend, struct hwloc_disc_status
       }
 
       /* get all memory info at once */
-      hwloc__levelzero_memory_get(dvh[j], osdev, nr_subdevices, subh, subosdevs);
+      hwloc__levelzero_memory_get(dvh[j], osdev, is_integrated, nr_subdevices, subh, subosdevs);
 
       parent = NULL;
-      res = zesDevicePciGetProperties(sdvh, &pci);
-      if (res == ZE_RESULT_SUCCESS) {
-        parent = hwloc_pci_find_parent_by_busid(topology,
-                                                pci.address.domain,
-                                                pci.address.bus,
-                                                pci.address.device,
-                                                pci.address.function);
-        if (parent && parent->type == HWLOC_OBJ_PCI_DEVICE) {
-          if (pci.maxSpeed.maxBandwidth > 0)
-            parent->attr->pcidev.linkspeed = ((float)pci.maxSpeed.maxBandwidth)/1000/1000/1000;
+#ifdef HWLOC_HAVE_ZEDEVICEPCIGETPROPERTIESEXT
+      { /* try getting PCI BDF+speed from core extension */
+        ze_pci_ext_properties_t ext_pci;
+        ext_pci.stype =  ZE_STRUCTURE_TYPE_PCI_EXT_PROPERTIES;
+        ext_pci.pNext = NULL;
+        res = zeDevicePciGetPropertiesExt(dvh[j], &ext_pci);
+        if (res == ZE_RESULT_SUCCESS) {
+          parent = hwloc_pci_find_parent_by_busid(topology,
+                                                  ext_pci.address.domain,
+                                                  ext_pci.address.bus,
+                                                  ext_pci.address.device,
+                                                  ext_pci.address.function);
+          if (parent && parent->type == HWLOC_OBJ_PCI_DEVICE) {
+            if (ext_pci.maxSpeed.maxBandwidth > 0)
+              parent->attr->pcidev.linkspeed = ((float)ext_pci.maxSpeed.maxBandwidth)/1000/1000/1000;
+          }
+        }
+      }
+#endif /* HWLOC_HAVE_LEVELZERO_CORE_PCI_EXT */
+      if (!parent) {
+        /* try getting PCI BDF+speed from sysman */
+        zes_pci_properties_t pci;
+        res = zesDevicePciGetProperties(sdvh, &pci);
+        if (res == ZE_RESULT_SUCCESS) {
+          parent = hwloc_pci_find_parent_by_busid(topology,
+                                                  pci.address.domain,
+                                                  pci.address.bus,
+                                                  pci.address.device,
+                                                  pci.address.function);
+          if (parent && parent->type == HWLOC_OBJ_PCI_DEVICE) {
+            if (pci.maxSpeed.maxBandwidth > 0)
+              parent->attr->pcidev.linkspeed = ((float)pci.maxSpeed.maxBandwidth)/1000/1000/1000;
+          }
         }
       }
       if (!parent)
